@@ -1,19 +1,16 @@
+"""Apps for managing kernel specs."""
 # Copyright (c) Jupyter Development Team.
 # Distributed under the terms of the Modified BSD License.
 import errno
 import json
 import os.path
 import sys
+import typing as t
 
-from jupyter_core.application import base_aliases
-from jupyter_core.application import base_flags
-from jupyter_core.application import JupyterApp
-from traitlets import Bool
-from traitlets import Dict
-from traitlets import Instance
-from traitlets import List
-from traitlets import Unicode
+from jupyter_core.application import JupyterApp, base_aliases, base_flags
+from traitlets import Bool, Dict, Instance, List, Unicode
 from traitlets.config.application import Application
+from traitlets.config.loader import Config
 
 from . import __version__
 from .kernelspec import KernelSpecManager
@@ -21,6 +18,8 @@ from .provisioning.factory import KernelProvisionerFactory
 
 
 class ListKernelSpecs(JupyterApp):
+    """An app to list kernel specs."""
+
     version = __version__
     description = """List installed kernel specifications."""
     kernel_spec_manager = Instance(KernelSpecManager)
@@ -42,11 +41,12 @@ class ListKernelSpecs(JupyterApp):
         return KernelSpecManager(parent=self, data_dir=self.data_dir)
 
     def start(self):
+        """Start the application."""
         paths = self.kernel_spec_manager.find_kernel_specs()
         specs = self.kernel_spec_manager.get_all_specs()
         if not self.json_output:
             if not specs:
-                print("No kernels available")
+                self.log.info("No kernels available")
                 return
             # pad to width of longest kernel name
             name_len = len(sorted(paths, key=lambda name: len(name))[-1])
@@ -60,14 +60,17 @@ class ListKernelSpecs(JupyterApp):
                 # not in jupyter path, artificially added to the front
                 return (-1, path)
 
-            print("Available kernels:")
+            self.log.info("Available kernels:")
             for kernelname, path in sorted(paths.items(), key=path_key):
-                print("  %s    %s" % (kernelname.ljust(name_len), path))
+                self.log.info("  %s    %s", kernelname.ljust(name_len), path)
         else:
-            print(json.dumps({"kernelspecs": specs}, indent=2))
+            self.log.info(json.dumps({"kernelspecs": specs}, indent=2))
+        return specs
 
 
 class InstallKernelSpec(JupyterApp):
+    """An app to install a kernel spec."""
+
     version = __version__
     description = """Install a kernel specification directory.
 
@@ -131,15 +134,17 @@ class InstallKernelSpec(JupyterApp):
     }
 
     def parse_command_line(self, argv):
+        """Parse the command line args."""
         super().parse_command_line(argv)
         # accept positional arg as profile name
         if self.extra_args:
             self.sourcedir = self.extra_args[0]
         else:
-            print("No source directory specified.")
+            self.log.info("No source directory specified.")
             self.exit(1)
 
     def start(self):
+        """Start the application."""
         if self.user and self.prefix:
             self.exit("Can't specify both user and prefix. Please choose one or the other.")
         try:
@@ -152,23 +157,21 @@ class InstallKernelSpec(JupyterApp):
             )
         except OSError as e:
             if e.errno == errno.EACCES:
-                print(e, file=sys.stderr)
+                self.log.exception(e)
                 if not self.user:
-                    print(
+                    self.log.warn(
                         "Perhaps you want to install with `sudo` or `--user`?",
-                        file=sys.stderr,
                     )
                 self.exit(1)
             elif e.errno == errno.EEXIST:
-                print(
-                    "A kernel spec is already present at %s" % e.filename,
-                    file=sys.stderr,
-                )
+                self.log.warn("A kernel spec is already present at %s", e.filename)
                 self.exit(1)
             raise
 
 
 class RemoveKernelSpec(JupyterApp):
+    """An app to remove a kernel spec."""
+
     version = __version__
     description = """Remove one or more Jupyter kernelspecs by name."""
     examples = """jupyter kernelspec remove python2 [my_kernel ...]"""
@@ -187,6 +190,7 @@ class RemoveKernelSpec(JupyterApp):
     flags.update(JupyterApp.flags)
 
     def parse_command_line(self, argv):
+        """Parse the command line args."""
         super().parse_command_line(argv)
         # accept positional arg as profile name
         if self.extra_args:
@@ -195,6 +199,7 @@ class RemoveKernelSpec(JupyterApp):
             self.exit("No kernelspec specified.")
 
     def start(self):
+        """Start the application."""
         self.kernel_spec_manager.ensure_native_kernel = False
         spec_paths = self.kernel_spec_manager.find_kernel_specs()
         missing = set(self.spec_names).difference(set(spec_paths))
@@ -202,9 +207,9 @@ class RemoveKernelSpec(JupyterApp):
             self.exit("Couldn't find kernel spec(s): %s" % ", ".join(missing))
 
         if not (self.force or self.answer_yes):
-            print("Kernel specs to remove:")
+            self.log.info("Kernel specs to remove:")
             for name in self.spec_names:
-                print("  %s\t%s" % (name.ljust(20), spec_paths[name]))
+                self.log.info("  %s\t%s", name.ljust(20), name.ljust(20))
             answer = input("Remove %i kernel specs [y/N]: " % len(self.spec_names))
             if not answer.lower().startswith("y"):
                 return
@@ -214,8 +219,8 @@ class RemoveKernelSpec(JupyterApp):
                 path = self.kernel_spec_manager.remove_kernel_spec(kernel_name)
             except OSError as e:
                 if e.errno == errno.EACCES:
-                    print(e, file=sys.stderr)
-                    print("Perhaps you want sudo?", file=sys.stderr)
+                    self.log.info(e, file=sys.stderr)
+                    self.log.info("Perhaps you want sudo?", file=sys.stderr)
                     self.exit(1)
                 else:
                     raise
@@ -223,11 +228,13 @@ class RemoveKernelSpec(JupyterApp):
 
 
 class InstallNativeKernelSpec(JupyterApp):
+    """An app to install the native kernel spec."""
+
     version = __version__
     description = """[DEPRECATED] Install the IPython kernel spec directory for this Python."""
     kernel_spec_manager = Instance(KernelSpecManager)
 
-    def _kernel_spec_manager_default(self):
+    def _kernel_spec_manager_default(self):  # pragma: no cover
         return KernelSpecManager(data_dir=self.data_dir)
 
     user = Bool(
@@ -247,23 +254,24 @@ class InstallNativeKernelSpec(JupyterApp):
         "debug": base_flags["debug"],
     }
 
-    def start(self):
+    def start(self):  # pragma: no cover
+        """Start the application."""
         self.log.warning(
             "`jupyter kernelspec install-self` is DEPRECATED as of 4.0."
             " You probably want `ipython kernel install` to install the IPython kernelspec."
         )
         try:
             from ipykernel import kernelspec
-        except ImportError:
-            print("ipykernel not available, can't install its spec.", file=sys.stderr)
+        except ModuleNotFoundError:
+            self.log.info("ipykernel not available, can't install its spec.", file=sys.stderr)
             self.exit(1)
         try:
             kernelspec.install(self.kernel_spec_manager, user=self.user)
         except OSError as e:
             if e.errno == errno.EACCES:
-                print(e, file=sys.stderr)
+                self.log.info(e, file=sys.stderr)
                 if not self.user:
-                    print(
+                    self.log.info(
                         "Perhaps you want to install with `sudo` or `--user`?",
                         file=sys.stderr,
                     )
@@ -272,22 +280,27 @@ class InstallNativeKernelSpec(JupyterApp):
 
 
 class ListProvisioners(JupyterApp):
+    """An app to list provisioners."""
+
     version = __version__
     description = """List available provisioners for use in kernel specifications."""
 
     def start(self):
+        """Start the application."""
         kfp = KernelProvisionerFactory.instance(parent=self)
-        print("Available kernel provisioners:")
+        self.log.info("Available kernel provisioners:")
         provisioners = kfp.get_provisioner_entries()
 
         # pad to width of longest kernel name
         name_len = len(sorted(provisioners, key=lambda name: len(name))[-1])
 
         for name in sorted(provisioners):
-            print(f"  {name.ljust(name_len)}    {provisioners[name]}")
+            self.log.info("  %s    %s", name.ljust(name_len), provisioners[name])
 
 
 class KernelSpecApp(Application):
+    """An app to manage kernel specs."""
+
     version = __version__
     name = "jupyter kernelspec"
     description = """Manage Jupyter kernel specifications."""
@@ -309,13 +322,18 @@ class KernelSpecApp(Application):
         }
     )
 
-    aliases = {}
-    flags = {}
+    aliases: t.Dict[t.Union[str, t.Tuple[str, ...]], t.Union[str, t.Tuple[str, str]]] = {}
+    flags: t.Dict[
+        t.Union[str, t.Tuple[str, ...]], t.Tuple[t.Union[t.Dict[str, t.Any], Config], str]
+    ] = {}
 
     def start(self):
+        """Start the application."""
         if self.subapp is None:
-            print("No subcommand specified. Must specify one of: %s" % list(self.subcommands))
-            print()
+            self.log.info(
+                "No subcommand specified. Must specify one of: %s" % list(self.subcommands)
+            )
+            self.log.info()
             self.print_description()
             self.print_subcommands()
             self.exit(1)
